/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.agents;

import edu.uiuc.ge.illigal.gale.config.GALEConfiguration;
import java.io.Serializable;

/** 
 * Implements the AgentPerformance for batch learning.
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha 
 * @since   0.9alpha 
 *
 */

public class BatchPerformance
implements AgentPerformance, Serializable, Cloneable
{
	//-------------------------------------------------------------------------
	
	/** Confusion matrix */
	protected int [][] imConfusion;

	/** Number of unclassified classified instances */
	protected int iDontKnow;

	/** Number of classified instances */
	protected int iTotal;

	/** Number of classes */
	protected int iCls;

	//-------------------------------------------------------------------------

	/**
	 * Initializes the performance object using the given configuration.
	 *
	 * @param galeCnf The GALE environment configuration
	 */
	public BatchPerformance ( GALEConfiguration galeCnf )
	{
		iCls = galeCnf.getNumClasses();
		imConfusion = new int[iCls][iCls];

		reset();
	}
 
	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public int [][] getConfusionMatrix ()
	{
		return imConfusion;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public int getDontKnow ()
	{
		return iDontKnow;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public int getTotal ()
	{
		return iTotal;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public float getAccuracy ()
	{
		float fAcc = 0;

		for ( int i=0 ; i<iCls ; i++ )
			fAcc += imConfusion[i][i];

		fAcc = (iTotal==0)?0:fAcc/iTotal;

		return fAcc;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public float getScaledAccuracy ()
	{
		float fTmp = getAccuracy();
	
		return fTmp*fTmp;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public void update ( int iCls, int iReal )
	{
		if ( iCls==-1 )
			iDontKnow++;
		else
			imConfusion[iCls][iReal]++;

		iTotal++;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public void reset ()
	{
		for ( int i=0 ; i<iCls ; i++ )
			for ( int j=0 ; j<iCls ; j++ )
				imConfusion[i][j] = 0;

		iDontKnow = iTotal = 0;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public boolean equals ( Object obj )
	{
		boolean bRes = false;

		if ( getClass().getName().equals(obj.getClass().getName()) ) {
			bRes = true;
			BatchPerformance bp = (BatchPerformance)obj;
			for ( int i=0 ; bRes && i<iCls ; i++ )
				for ( int j=0 ; bRes && j<iCls ; j++ )
					if ( imConfusion[i][j]!=bp.imConfusion[i][j] )
						bRes = false;
		}
		
		return bRes;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public Object clone()
	throws CloneNotSupportedException
	{
		BatchPerformance bp = (BatchPerformance)super.clone();

		bp.iCls        = iCls;
		bp.iTotal      = iTotal;
		bp.imConfusion = new int[iCls][iCls];

		for ( int i=0 ; i<iCls ; i++ )
			for ( int j=0 ; j<iCls ; j++ )
				bp.imConfusion[i][j]=imConfusion[i][j];


		return bp;
	}

	//-------------------------------------------------------------------------

	// Agent Performance implementation
	public String toString()
	{
		String sRes = "";
		
		sRes += "Model Performance\n";
		sRes += "----------------\n";
		float fTmp = getAccuracy()*getTotal();
		sRes += "Correct         : "+(int)fTmp+"\n";
		sRes += "Incorrect       : "+(int)(getTotal()-getDontKnow()-fTmp)+"\n";
		sRes += "Don't know      : "+getDontKnow()+"\n";
		sRes += "Total           : "+getTotal()+"\n";
		sRes += "Accuracy        : "+getAccuracy()+"\n";
		if ( getAccuracy()>0 ) {
			fTmp /= getTotal()-getDontKnow();
		}
		sRes += "Action Accuracy : "+fTmp+"\n";
		sRes += "Scaled Accuracy : "+getScaledAccuracy()+"\n\n";

		sRes += "Confusion Matrix (Predicted\\Real)\n";
		sRes += "----------------------------------\n";
		for ( int i=0 ; i<iCls ; i++ ) {
			for ( int j=0 ; j<iCls ; j++ )
				sRes += imConfusion[i][j]+"\t";
			sRes += "\n";
		}

		return sRes;
	}

	//-------------------------------------------------------------------------

}
