/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.agents;

import edu.uiuc.ge.illigal.gale.config.*;
import edu.uiuc.ge.illigal.gale.data.*;
import edu.uiuc.ge.illigal.gale.distance.*;

import java.io.Serializable;
import java.util.Vector;
import java.util.Random;

/** 
 * Implements the generic handle to Agents for Induction Trees.
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha 
 * @since   0.9alpha 
 *
 */

public class DTAgent
implements Agent, Serializable, Cloneable
{
	//-------------------------------------------------------------------------

	/** The GALE environment configuration object */
   protected GALEConfiguration galeCnf;

	/** The attributes set information object */
   protected AttributeSet attCnf;

   /** The threshold for real attribute classification */
   protected float fThres;

   /** The attribute used */
   protected Attribute aAtt;

	/** The index of the attribute used */
	protected int iIndex;

	/** The class if the node is a leaf */
	protected int iCls;

   /** The Agent's children */
   protected Vector vecChild;

   /** Agent Performance object*/
	protected AgentPerformance agnPer;

	/** Random number generator object */
	protected Random rnd;

	//------------------------------------------------------------------------

   /** 
	 * Builds a simple agent with null information
    */ 
      
   public DTAgent ()
   {
		attCnf  = null;
   	galeCnf = null;
		rnd     = null;

      fThres   = 0;
      iCls     = 0;
      iIndex   = 0;
      aAtt     = null;
      vecChild = null;
		agnPer   = null;
   }
   
	//------------------------------------------------------------------------

   /** 
	 * Builds a simple agent. The Agent is build using the GALE configuration
    * object.
	 *
    * @param gale The GALE configuration
	 * @param att The instance data information used
    */ 
      
   public DTAgent ( GALEConfiguration gale, AttributeSet att )
   {
      int i, iMax;
		int iAtt = gale.getNumAttributes();	

		attCnf  = att;
   	galeCnf = gale;
		rnd     = gale.getRandom();
      iIndex  = rnd.nextInt(iAtt);
		aAtt    = attCnf.attribute(iIndex);
		iCls    = rnd.nextInt(galeCnf.getNumClasses());
		
		if ( aAtt.type()==Attribute.REAL || aAtt.type()==Attribute.INTEGER ) {
			fThres = boundedFloat(aAtt.minimum(), aAtt.maximum());
		}
		else 
			fThres = 0;

      vecChild = new Vector();

		agnPer = new BatchPerformance(galeCnf);
   }
   
	//---------------------------------------------------------------------
	
	// Agent implementation
   public Agent buildAgent ( GALEConfiguration galeCnf, 
	                                    AttributeSet att )
	{
		rnd = galeCnf.getRandom();

		return buildDTAgent ( galeCnf,
		                       att,
		                       rnd.nextInt(galeCnf.getAgentMaxInitDepth()-1)+1
	                        );
	}

	//---------------------------------------------------------------------
	
   /** 
	 * Builds an Agent Hierarchical Aggregate.
 	 *
	 * @param galeCnf The GALE environment configuration object
	 * @param attSet The attribute set information object
	 * @param iDepth The Agent depth
	 * @return The Random Agent build
	 */
   protected static DTAgent buildDTAgent ( 
                                       GALEConfiguration galeCnf, 
	                                    AttributeSet attSet,
	                                    int iDepth 
                                      )
   {
		Random rnd = galeCnf.getRandom();
      DTAgent agnRes = new DTAgent(galeCnf, attSet);

      if ( iDepth>0 ) {
			if ( agnRes.aAtt.type()==Attribute.NOMINAL ) {
				for ( int i=0, iMax=(int)agnRes.aAtt.maximum() ; i<iMax ; i++ ) {
					agnRes.addChild(
						DTAgent.buildDTAgent(galeCnf,attSet,iDepth-1)
					);
				}
			}
			else if ( agnRes.aAtt.type()==Attribute.INTEGER || 
			          agnRes.aAtt.type()==Attribute.REAL ){
				agnRes.addChild(
					DTAgent.buildDTAgent(galeCnf,attSet,iDepth-1)
				);
				agnRes.addChild(
					DTAgent.buildDTAgent(galeCnf,attSet,iDepth-1)
				);
			}
      }

      return agnRes;
   }

	//-------------------------------------------------------------------------

	// Agent implementation
	public AgentPerformance getPerformance ()
	{
		return agnPer;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public void resetPerformance ()
	{
		agnPer.reset();

		for ( int i=0, iMax=getChildNumber(); i<iMax ; i++ )
			getChild(i).resetPerformance();
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent merge ( Agent agn )
	{
		DTAgent agnRes = null;
		int     icn1   = getChildNumber();
		int     icn2   = ((DTAgent)agn).getChildNumber();

		try {
			if ( icn1==0 || icn2==0 ) {
				if ( icn1>0 )                 agnRes = (DTAgent)this.clone();
				else if ( icn2>0 )            agnRes = (DTAgent)agn.clone();
				else if ( rnd.nextBoolean() ) agnRes = (DTAgent)this.clone();
				else                          agnRes = (DTAgent)agn.clone();
			}
			else {
				// Crossover operator
				DTAgent agn1 = (DTAgent)this.clone();
				DTAgent agn2 = (DTAgent)agn.clone();
		
				DTAgent agnNode1 = null;
				DTAgent agnNode2 = null;

				do {
					int iNode1 = rnd.nextInt(agn1.countNodes());
					agnNode1 = getIthNode(agn1,iNode1);
//System.out.println("* "+iNode1+","+agnNode1.getChildNumber()+" *");
				} while ( agnNode1.getChildNumber()==0 );
//System.out.println(agnNode1.getStringModel(""));

				do {
					int iNode2 = rnd.nextInt(agn2.countNodes());
					agnNode2 = getIthNode(agn2,iNode2);
//System.out.println("* "+iNode2+","+agnNode2.getChildNumber()+" *");
				} while ( agnNode2.getChildNumber()==0 );
//System.out.println(agnNode2.getStringModel(""));

				agnNode1.swapChildren(agnNode2);

				if ( rnd.nextBoolean() )
					agnRes = agn1;
				else
					agnRes = agn2;
			}
		}
		catch ( Exception e ) {
			System.err.println("Failure in merge for "+getClass().getName());
			e.printStackTrace();
			System.exit(1);
		}

		return agnRes;
	}

	//------------------------------------------------------------------------

	/**
	 * Exchange children
	 *
	 * @param agn The other DTAgent to exchange children 
	 */
      
   protected void swapChildren ( DTAgent agn )
   {
		int icn1 = getChildNumber();
		int icn2 = agn.getChildNumber();

		if ( icn1>0 && icn2>0 ) {
			int iAgn1 = rnd.nextInt(icn1);
			int iAgn2 = rnd.nextInt(icn2);

			DTAgent agnTmp = getChild(iAgn1);
			addChild(agn.getChild(iAgn2),iAgn1);
			agn.addChild(agnTmp,iAgn2);
		}
   }

	//------------------------------------------------------------------------

   /** 
	 * Returns the ith agent in the tree
	 *
    * @return the ith agent in breadth first search
    */ 
	protected DTAgent getIthNode ( DTAgent agn, int iNode )
	{
		DTAgent agnTmp = null;;
		Vector vecNodes = new Vector();

		vecNodes.addElement(agn);
		while ( iNode>0 ) {
			agnTmp = (DTAgent)vecNodes.elementAt(0);
			vecNodes.removeElementAt(0);
			for ( int i=0, iMax=agnTmp.getChildNumber() ; i<iMax ; i++ )
				vecNodes.addElement(agnTmp.getChild(i));
			iNode--;
		}

		return (DTAgent)vecNodes.elementAt(0);
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent split ()
	{
		DTAgent agn = null;

		try {
			agn = (DTAgent)clone();
			agn.mutateGenotype();
		}
		catch ( Exception e ) {
			System.err.println("Failure in split for "+getClass().getName());
			e.printStackTrace();
			System.exit(1);
		}
			
		return agn;
	}

	//-------------------------------------------------------------------------

	/**
    * Mutates the genetic information contained in the tree
	 */
	protected void mutateGenotype ()
	{
		float fPM = galeCnf.getSomaticMP();

		if ( rnd.nextFloat()<fPM ) 
			iCls = rnd.nextInt(galeCnf.getNumClasses());
		
		if ( aAtt.type()==Attribute.REAL || aAtt.type()==Attribute.INTEGER ) 
			if ( rnd.nextFloat()<fPM ) 
				fThres = boundedFloat(aAtt.minimum(), aAtt.maximum());

		if ( rnd.nextFloat()<fPM && getChildNumber()>0 ) {
			int  iNewIndex = rnd.nextInt(galeCnf.getNumAttributes());
			Attribute aNew = attCnf.attribute(iNewIndex);

			if ((aAtt.type()==Attribute.REAL||aAtt.type()==Attribute.INTEGER)&&
			    (aNew.type()==Attribute.REAL||aNew.type()==Attribute.INTEGER)){
				// Both are numeric
				iIndex = iNewIndex;
				aAtt   = aNew;
				for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) 
					getChild(i).mutateGenotype();
			}
			else {
				int iONC = -1;
				int iNNC = -1;

				// Compute the number of children
				if (aAtt.type()==Attribute.REAL||aAtt.type()==Attribute.INTEGER) 
					iONC = 2;
				else if ( aAtt.type()==Attribute.NOMINAL )
					iONC = (int)aAtt.maximum(); 

				if (aNew.type()==Attribute.REAL||aNew.type()==Attribute.INTEGER) 
					iNNC = 2;
				else if ( aNew.type()==Attribute.NOMINAL )
					iNNC = (int)aNew.maximum(); 

				// Possible change on the size of children vector
				if ( iONC>iNNC ) {
					// Too much children
					for ( int i=iNNC, iMax=iONC ; i<iMax ; i++ )
						removeChild(iNNC);

					for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) 
						getChild(i).mutateGenotype();
				}
				else if ( iONC<iNNC ) {
					// Too few children
					for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) 
						getChild(i).mutateGenotype();
					for ( int i=iONC, iMax=iNNC ; i<iMax ; i++ )
						addChild( 
						         buildDTAgent ( galeCnf,
		                                    attCnf,
		                       rnd.nextInt(galeCnf.getAgentMaxInitDepth()-1)+1 )
	                       );
				}
				iIndex = iNewIndex;
				aAtt   = aNew;
			} 
		}
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Object clone()
	throws CloneNotSupportedException
	{
		DTAgent nag = (DTAgent)super.clone();

      nag.galeCnf   = galeCnf;
		nag.attCnf    = attCnf;
		nag.aAtt      = aAtt;
		nag.iIndex    = iIndex;
		nag.rnd       = rnd;
      nag.fThres    = fThres;
      nag.iCls      = iCls;
      nag.vecChild  = (Vector)vecChild.clone();
		nag.agnPer    = (AgentPerformance)agnPer.clone();

      for ( int i=0, iMax=nag.vecChild.size() ; i<iMax ; i++ )
         nag.addChild((DTAgent)nag.getChild(i).clone(), i);
 
      return nag;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public float complexity ()
	{
		return countNodes();
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent prune ()
	{
		Agent agnRes = this;

		if ( getChildNumber()>0 ) {
			if ( aAtt.type()==Attribute.NOMINAL ) {
//System.out.println("*NOMINAL node*");
				int iChilds = getChildNumber();
				int inumGG  = 0;
				int inumZG  = 0;
				int inumZZ  = 0;
				DTAgent agn = null;

				// Computing information for deciding the prune case
				for ( int i=0 ; i<iChilds ; i++ ) {
					agn = getChild(i);
					int   iTot = agn.getPerformance().getTotal();
					float fAcc = agn.getPerformance().getAccuracy();

					if ( fAcc>0 && iTot>0 )       inumGG++;
					else if ( fAcc==0 && iTot>0 ) inumZG++;
					else                          inumZZ++;
				}
				if ( inumGG==iChilds ) {
					// Case 1
//System.out.println("==> Case 1 ("+iChilds+","+inumGG+","+inumZG+","+inumZZ+")");
					agnRes = this;
					for ( int i=0 ; i<iChilds ; i++ ) 
						addChild((DTAgent)getChild(i).prune(),i);
				}
				else if ( inumGG>1 ) {
					// Case 2
//System.out.println("==> Case 2 ("+iChilds+","+inumGG+","+inumZG+","+inumZZ+")");
					agnRes = this;
					for ( int i=0 ; i<iChilds ; i++ ) {
						agn = getChild(i);
						int   iTot = agn.getPerformance().getTotal();
						float fAcc = agn.getPerformance().getAccuracy();

						if ( fAcc>0 ) {
							addChild((DTAgent)agn.prune(),i);
						}
						else {
							agn.removeAllChilds();
						}
					}
				}
				else if ( inumGG==1 ) {
					// Case 3
//System.out.println("==> Case 3 ("+iChilds+","+inumGG+","+inumZG+","+inumZZ+")");
					agnRes = this;

					boolean bStop = false;
					for ( int i=0 ; i<iChilds && !bStop ; i++ ) {
						agn = getChild(i);
						float fAcc = agn.getPerformance().getAccuracy();
						if ( fAcc>0 ) bStop = true;
					}

					galeCnf   = agn.galeCnf;
					attCnf    = agn.attCnf;
		      	fThres    = agn.fThres;
					aAtt      = agn.aAtt;
					iIndex    = agn.iIndex;
					rnd       = agn.rnd;
		      	iCls      = agn.iCls;
		      	vecChild  = agn.vecChild;
					//agnPer    = agn.agnPer;
		
					agnRes = prune();
				}
				else {
//System.out.println("==> Case 4 ("+iChilds+","+inumGG+","+inumZG+","+inumZZ+")");
					// Case 4
					removeAllChilds();
					agnRes = this;
				}
			}
			else if ( aAtt.type()==Attribute.REAL ||
		     			 aAtt.type()==Attribute.INTEGER ) {
//System.out.println("*NUMERIC node("+getChildNumber()+")*");
				DTAgent agnLE = getChild(0);
				DTAgent agnGR = getChild(1);

		   	float faccLE = agnLE.getPerformance().getAccuracy();
		   	float faccGR = agnGR.getPerformance().getAccuracy();
				int   itotLE = agnLE.getPerformance().getTotal();
				int   itotGR = agnGR.getPerformance().getTotal();
				
				if ( faccLE>0 && faccGR>0 ) {
//System.out.println("==> Case 1 ("+faccLE+","+itotLE+","+faccGR+","+itotGR+")");
					// Case 1
					agnRes = this;
					addChild((DTAgent)agnLE.prune(),0);
					addChild((DTAgent)agnGR.prune(),1);
				}
				else if ( faccLE==0 && itotLE>0 &&
			          	faccGR>0  && itotGR>0 ) {
//System.out.println("==> Case 2a ("+faccLE+","+itotLE+","+faccGR+","+itotGR+")");
					// Case 2a
					agnRes = this;
					agnLE.removeAllChilds();
					addChild((DTAgent)agnGR.prune(),1);
				}
				else if ( faccLE>0  && itotLE>0 &&
			          	faccGR==0 && itotGR>0 ) {
//System.out.println("==> Case 2b ("+faccLE+","+itotLE+","+faccGR+","+itotGR+")");
					// Case 2b
					agnRes = this;
					agnGR.removeAllChilds();
					addChild((DTAgent)agnLE.prune(),0);
				}
				else if ( faccLE==0 && itotLE==0 &&
			          	faccGR>0  && itotGR>0 ) {
//System.out.println("==> Case 3a ("+faccLE+","+itotLE+","+faccGR+","+itotGR+")");
					// Case 3a
					galeCnf   = agnGR.galeCnf;
					attCnf    = agnGR.attCnf;
		      	fThres    = agnGR.fThres;
					aAtt      = agnGR.aAtt;
					iIndex    = agnGR.iIndex;
					rnd       = agnGR.rnd;
		      	iCls      = agnGR.iCls;
		      	vecChild  = agnGR.vecChild;
					//agnPer    = agnGR.agnPer;
		
					agnRes = prune();
				}
				else if ( faccLE>0  && itotLE>0 &&
			          	faccGR==0 && itotGR==0 ) {
//System.out.println("==> Case 3b ("+faccLE+","+itotLE+","+faccGR+","+itotGR+")");
					// Case 3b
					galeCnf   = agnLE.galeCnf;
					attCnf    = agnLE.attCnf;
		      	fThres    = agnLE.fThres;
					aAtt      = agnLE.aAtt;
					iIndex    = agnLE.iIndex;
					rnd       = agnLE.rnd;
		      	iCls      = agnLE.iCls;
		      	vecChild  = agnLE.vecChild;
					agnPer    = agnLE.agnPer;
	
					agnRes = prune();
				}
				else {
//System.out.println("==> Case 4 ("+faccLE+","+itotLE+","+faccGR+","+itotGR+")");
					// Case 4
					removeAllChilds();
					agnRes = this;
				}
			}
		}
	
		return agnRes;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public int classify ( Instance ins )
	{
		int iInsCls = ins.getInstanceClass();
		int iResCls = -1;

		if ( getChildNumber()==0 ) {
			iResCls = iCls;
		}
		else {
			float fVal = ins.getValue(iIndex);

			if ( !ins.isKnown(iIndex) ) {
				iResCls = -1;
			}
			else {
				if ( aAtt.type()==Attribute.REAL ||
			     	aAtt.type()==Attribute.INTEGER ) {
					if ( fVal<=fThres ) {
						iResCls = getChild(0).classify(ins);
					}
					else {
						iResCls = getChild(1).classify(ins);
					}
				}
				else if ( aAtt.type()==Attribute.NOMINAL ) {
					iResCls = getChild((int)fVal).classify(ins);
				}
			}
		}
		
		agnPer.update(iResCls,iInsCls);
		return iResCls;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public String toString()
	{
		String sRes = "Decision Tree Model\n";
		sRes += "-------------------\n\n";
	
		sRes += "Number of Nodes  :"+countNodes()+"\n";
		sRes += "Number of Leaves :"+countLeafNodes()+"\n\n";
		sRes += "*DT* {"+agnPer.getAccuracy()+","+agnPer.getDontKnow()+",";
		sRes += agnPer.getTotal()+"}\n";


		return sRes+getStringModel("")+"\n\n"+agnPer;
	}

	//-------------------------------------------------------------------------

	/**
    * Returns the stringfied model for the agent.
	 *
	 * @param pre the current prefix
	 * @return the stringfied model
	 */
	protected String getStringModel (String pre )
	{
		String    sRes = "";
		Attribute cls = attCnf.attribute(attCnf.size()-1);
	
		if ( getChildNumber()==0 ) {
			// Tree with one node
			sRes += pre+"*Always* : "+cls.tag(iCls)+" ";	
			sRes += " {"+agnPer.getAccuracy()+","+agnPer.getDontKnow()+",";
			sRes += agnPer.getTotal()+"}\n";
		}
		else {
			// Other cases
			if ( aAtt.type()==Attribute.NOMINAL ) {
				for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) {
					// Nominal child
					DTAgent agn = getChild(i);
					sRes += pre+aAtt.name()+"="+aAtt.tag(i);
					if ( agn.getChildNumber()==0 ) {
						sRes += " : "+cls.tag(agn.iCls)+" ";
					}
					sRes += " {"+agn.agnPer.getAccuracy()+","+agn.agnPer.getDontKnow()+",";
					sRes += agn.agnPer.getTotal()+"}\n";
					if ( agn.getChildNumber()>0 ) {
						sRes += agn.getStringModel(pre+"|\t");
					}
				}
			}
			else if ( aAtt.type()==Attribute.REAL ||
			          aAtt.type()==Attribute.INTEGER ) {
				// Less or equal child
				DTAgent agn = getChild(0);
				sRes += pre+aAtt.name()+"<="+fThres;
				if ( agn.getChildNumber()==0 ) {
					sRes += " : "+cls.tag(agn.iCls)+" ";
				}
				sRes += " {"+agn.agnPer.getAccuracy()+","+agn.agnPer.getDontKnow()+",";
				sRes += agn.agnPer.getTotal()+"}\n";
				if ( agn.getChildNumber()>0 ) {
					sRes += agn.getStringModel(pre+"|\t");
				}
				// Greater child
				agn = getChild(1);
				sRes += pre+aAtt.name()+">"+fThres;
				if ( agn.getChildNumber()==0 ) {
					sRes += "  : "+cls.tag(agn.iCls)+" ";
				}
				sRes += " {"+agn.agnPer.getAccuracy()+","+agn.agnPer.getDontKnow()+",";
				sRes += agn.agnPer.getTotal()+"}\n";
				if ( agn.getChildNumber()>0 ) {
					sRes += agn.getStringModel(pre+"|\t");
				}
			}
		}

		return sRes;
	}

	//-------------------------------------------------------------------------

	/**
    * Returns the star tree.
	 *
	 * @return the stringfied star tree
	 */
	protected String getStarTree()
	{
		String sRes = "(*";

		for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) 
		   sRes += ((DTAgent)vecChild.elementAt(i)).getStarTree();

		return sRes+")";
	}

	//-------------------------------------------------------------------------

	/**
    * Generates a random bounded float using the random number generator.
	 *
	 * @param lb lower bound
	 * @param ub upper bound
	 */
	protected float boundedFloat ( float lb, float ub )
	{
		float fDiff = ub-lb;

		fDiff = rnd.nextFloat()*fDiff;

		return fDiff+lb;
	}
	
	//-------------------------------------------------------------------------


   /** 
	 * Adds a new Agent child. Links the agnChild to the current agent.
	 *
    * @param agnChild The Agent to be added
    */ 
   protected void addChild ( DTAgent agnChild ) 
   {
		vecChild.addElement(agnChild);
   }

   //------------------------------------------------------------------------

   /** 
	 * Adds a new Agent child. Links the agnChild to the current agent.
	 *
    * @param agnChild The Agent to be added
    * @param iPos The child position to be added
    */ 
   protected void addChild ( DTAgent agnChild, int iPos ) 
   {
		vecChild.setElementAt(agnChild, iPos);
   }

   //------------------------------------------------------------------------

   /** 
	 * Remove the ith Child.
	 *
    * @param iPos The child position to be removed
    */ 
   protected void removeChild ( int iPos ) 
   {
		vecChild.removeElementAt(iPos);
   }

	//------------------------------------------------------------------------

   /** 
	 * Remove all children
    */ 
   protected void removeAllChilds () 
   {
		vecChild.removeAllElements();
   }

   //------------------------------------------------------------------------

   /** 
	 * Returns the requested child. Returns the i child of the Agent.
	 *
    * @param iChild The requested Agent child
    * @return The Agent child
    */ 
   protected DTAgent getChild ( int iChild ) 
   {
      return (DTAgent)vecChild.elementAt(iChild);
   }

  	//------------------------------------------------------------------------

   /** 
	 * Returns the number of agent childs. 
	 *
    * @return The Agent child number
    */ 
   protected int getChildNumber ()
   {
      return vecChild.size();
   }

	//------------------------------------------------------------------------

   /** 
	 * Count the number of nodes in the tree
	 *	
	 * @return The number of nodes
	 */
   protected int countNodes ()
   {
      int iRes = 1;
	
      for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) 
			iRes += getChild(i).countNodes();

		return iRes;
   }

	//------------------------------------------------------------------------

   /** 
	 * Count the number of leaf agents in the tree 
	 *	
	 * @return The number of leaves in the tree
	 */
   protected int countLeafNodes ()
   {
		if ( getChildNumber()==0 )
			return  1;
		else {
      	int iRes = 0;
      	for ( int i=0, iMax=getChildNumber() ; i<iMax ; i++ ) 
				iRes += getChild(i).countLeafNodes();

			return iRes;
		}
   }


	//-------------------------------------------------------------------------

	public static void main ( String [] sArgs )
	{
		try {
			GALEConfiguration gg = new FileGALEConfiguration("gale."+sArgs[0]+".cnf");
			InstanceSet       ii = new FileInstanceSet(gg,sArgs[0]+".arff");

			Agent a = new DTAgent().buildAgent(gg,ii.attributeSet());
			Agent b = new DTAgent().buildAgent(gg,ii.attributeSet());

			for ( int i=0, iMax=ii.getSize() ; i<iMax ; i++ ) {
				a.classify(ii.getInstance(i));	
				b.classify(ii.getInstance(i));	
			}

			System.out.println("Pruning the induced tree");
			System.out.println("-------------------------\n");
			a=a.prune();
			b=b.prune();
			a.resetPerformance();
			b.resetPerformance();
			for ( int i=0, iMax=ii.getSize() ; i<iMax ; i++ ) {
				a.classify(ii.getInstance(i));	
				b.classify(ii.getInstance(i));	
			}
			System.out.println(a);
			System.out.println(b);
			System.out.println("Crossing the induced trees");
			System.out.println("--------------------------\n");
			DTAgent c = (DTAgent)a.merge(b);
			c.resetPerformance();
			for ( int i=0, iMax=ii.getSize() ; i<iMax ; i++ ) {
				c.classify(ii.getInstance(i));	
			}
			System.out.println(c);

			System.out.println("Splitting the induced trees");
			System.out.println("---------------------------\n");
			DTAgent d = (DTAgent)c.split();
			d.resetPerformance();
			for ( int i=0, iMax=ii.getSize() ; i<iMax ; i++ ) {
				d.classify(ii.getInstance(i));	
			}
			System.out.println(d);
		
		}
		catch ( Exception e ) {
			e.printStackTrace();
		}
	}

	//------------------------------------------------------------------------

}
