/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.config;

import edu.uiuc.ge.illigal.gale.distance.Distance;
import edu.uiuc.ge.illigal.gale.agents.Agent;

import java.util.Random;
import java.io.Serializable;

/**
 * This <strong>abstract</strong> class contains the access to the
 * GA configuration object
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha
 * @since 0.9alpha
 */

public abstract class GALEConfiguration 
implements Serializable
{
   //---------------------------------------------------------------------
	
	/** Test constant for using the best */
	public final static int TEST_BEST = 0;

	/** Test constant for using bagging */
	public final static int TEST_BAGGING = 1;

   //---------------------------------------------------------------------

   /** Number of problem features */
   protected int iNumFeatures;

   /** Number of possible target classes */
   protected int iNumClasses;

   /** Random Number Generator */
   protected Random rndGen;

   /** Random Number Generator seed */
   protected long lSeed;
   
   /** Random Number Generator seed for board runs*/
   protected long lSeedRuns;
   
   /** The number of initial features used by the agent */
   protected int iAgentInitFeatures;

	/** The number of world iterations */
	protected int iWorldIterations;

   /** Somatic Mutation Probability */
   protected float fSomaticMP;

   /** Merge Probability */
   protected float fMergeP;

   /** KThreshold for survival pressure */
   protected float fKThreshold;

   /** Maximal split probability */
   protected float fMaxSplitP;

   /** Maximal initial agent width */
   protected int iAgentMaxInitWidth;

   /** Maximal initial agent depth */
   protected int iAgentMaxInitDepth;

	/** Initial aggregates number */
   protected float fInitialAggregates;

	/** Board X width */
   protected int iMaxX;

	/** Board Y height */
   protected int iMaxY;

	/** The mapping allocation strategy */
	protected String sMappingStrategy;

   /** Random picking proportion */
   protected float fProportion;

	/** Check if pruning is activates */
	protected boolean bPrune;
	
	/** Test mode. 
	 * <ul>
	 * 	<li> TEST_BEST: test using the best
	 * 	<li> TEST_BAGGING: test using bagging
	 * </ul>
	 */
	protected int iTestMode;

	/** Distance function to use */
	protected Distance dis;
	
	/** Agents to use in the run */
	protected Agent aAgent;
	
   //---------------------------------------------------------------------

   /** Returns the number of features that describes the problem.
	 *
    *  @return The number of features
    */
   public int getNumAttributes()
   {
      return iNumFeatures;
   }

   //---------------------------------------------------------------------

   /** Returns the number of target classes for the problem.
	 *
    *  @return The number of classes
    */
   public int getNumClasses()
   {
      return iNumClasses;
   }

	//---------------------------------------------------------------------

   /** Returns the random number generator.
	 *
    *  @return the random number generator
    */
   public Random getRandom()
   {
      return rndGen;
   }


   //---------------------------------------------------------------------

   /** Returns the initial random seed
	 *
    *  @return The initial random seed
    */
   public long getSeed()
   {
      return lSeed;
   }
	
	//---------------------------------------------------------------------

   /** Returns the initial random seed
	 *
    *  @return The initial random seed
    */
   public long getSeedRuns()
   {
      return lSeedRuns;
   }

   //---------------------------------------------------------------------

   /** Returns the number of initial features used.
	 *
    *  @return The number of initial features
    */
   public int getAgentInitAttributes()
   {
      return iAgentInitFeatures;
   }

	//---------------------------------------------------------------------

   /** Returns the number of the world iterations 
	 *
    *  @return The world iteration numbers
    */
   public int getWorldIterations ()
   {
      return iWorldIterations;
   }

	//---------------------------------------------------------------------

   /** Returns the somatic mutation probability.
	 *
    *  @return The somatic mutation probability
    */
   public float getSomaticMP ()
   {
      return fSomaticMP;
   }

	//---------------------------------------------------------------------

   /** Returns the merge probability.
	 *
    *  @return The merge probability
    */
   public float getMergeP ()
   {
      return fMergeP;
   }
 
   //---------------------------------------------------------------------
 
   /** Returns the extinctive kThreshold pressure.
	 *
    *  @return The kThreshold pressure
    */
   public float getKThreshold ()
   {
      return fKThreshold ;
   }

   //---------------------------------------------------------------------
 
   /** Returns the maximal Split Probability.
	 *
    *  @return The split probability
    */
   public float getMaxSplitP ()
   {
      return fMaxSplitP;
   }

   //---------------------------------------------------------------------

   /** Returns the agent initial maximal width.
	 *
    *  @return The maximal width
    */
   public int getAgentMaxInitWidth()
   {
      return iAgentMaxInitWidth;
   }

   //---------------------------------------------------------------------

   /** Returns the agent initial maximal depth.
	 *
    *  @return The maximal depth
    */
   public int getAgentMaxInitDepth()
   {
      return iAgentMaxInitDepth;
   }

	//---------------------------------------------------------------------

   /** Returns the initial aggregates present in the board.
	 *
    *  @return The number of aggregates
    */
   public float getInitialAggregates()
   {
      return fInitialAggregates;
   }

	//---------------------------------------------------------------------

   /** Returns the proportion of instances to pick randomly.
	 *
    *  @return The proportion of instances used
    */
   public float getInstancesProportion()
   {
      return fProportion;
   }

   //---------------------------------------------------------------------

   /** Returns the board X width.
	 *
    *  @return The width in cells
    */
   public int getMaxX()
   {
      return iMaxX;
   }

	//---------------------------------------------------------------------

   /** Returns the board Y height.
	 *
    *  @return The height in cells
    */
   public int getMaxY()
   {
      return iMaxY;
   }

	//---------------------------------------------------------------------

   /** Check for pruning.
	 *
    *  @return true if prune is required
    */
   public boolean getPrune()
   {
      return bPrune;
   }

	//---------------------------------------------------------------------

   /** Returns the test mode.
	 *
    *  @return the test mode
    */
   public int getTestMode()
   {
      return iTestMode;
   }


	//---------------------------------------------------------------------

   /** Returns the resource allocation strategy.
	 *
    *  @return The name of the mapping strategy
    */
   public String getMappingStrategy()
   {
      return sMappingStrategy;
   }
	
	//---------------------------------------------------------------------

   /** Returns the distance function to use
	 *
    *  @return The distance function
    */
   public Distance getDistance()
   {
      return dis;
   }

	//---------------------------------------------------------------------

   /** Returns the distance function to use
	 *
    *  @return The distance function
    */
   public Agent getAgent()
   {
      return aAgent;
   }

   //---------------------------------------------------------------------

   /** Returns the GA Param Configuration as a String.

       @return The GA parameters configuration
    */
   public String toString()
   {
      String sRes = "";

      sRes += "#-----------------------------------------------------------------------------\n";
      sRes += "#\n";
      sRes += "# ALE Configuration file\n";
      sRes += "#\n";
      sRes += "#-----------------------------------------------------------------------------\n\n";

      sRes += "###\n";
      sRes += "### Problem Description\n";
      sRes += "###\n\n";

      sRes += "NumAttributes       = "+iNumFeatures+"\n";
      sRes += "NumClasses          = "+iNumClasses+"\n\n";

      sRes += "###\n";
      sRes += "### Agent's information\n";
      sRes += "###\n\n";
      sRes += "AgentInitAttributes = "+iAgentInitFeatures+"\n";
      sRes += "Prune               = "+bPrune+"\n";
      sRes += "TestMode            = "+
				  ((iTestMode==TEST_BEST)?"Best":"Bagging")
				  +"\n";
      sRes += "Distance            = "+dis+"\n\n";

      sRes += "###\n";
      sRes += "### GALE execution parameter\n";
      sRes += "###\n\n";

      sRes += "# General values\n";
      sRes += "WorldIterations = "+iWorldIterations+"\n\n";

      sRes += "# Board mapping\n";
      sRes += "MaxX = "+iMaxX+"\n";
      sRes += "MaxY = "+iMaxY+"\n\n";

		sRes += "# Initializations\n";
      sRes += "Seed              = "+lSeed+"\n";
      sRes += "InitialAggregates = "+fInitialAggregates+"\n";
      sRes += "AgentMaxInitWidth = "+iAgentMaxInitWidth+"\n";
      sRes += "AgentMaxInitDepth = "+iAgentMaxInitDepth+"\n\n";

		sRes += "# Parameters\n";	
      sRes += "MergeP           = "+fMergeP+"\n";
      sRes += "SomaticMutationP = "+fSomaticMP+"\n";
      sRes += "MaxSplitP        = "+fMaxSplitP+"\n";
      sRes += "kThreshold       = "+fKThreshold+"\n";

      return sRes;
   }

}
